package edu.uoc.main;

import java.io.IOException;
import java.util.HashMap;
import java.util.Map;

import edu.uoc.messages.AlarmMsg;

public class AlarmMsgHandler {

	private final static byte TEMPERATURE_AUTOMATIC = 0x05;
	private final static byte TEMPERATURE_MANUAL = 0x09;
	private final static byte TEMPERATURE = 0x01;
	private final static byte BATTERY = 0x02;
	
	private final static byte ALARM_SENT = 0x10;
	private final static byte ALARM_SW_ACKED = 0x20;
	private final static byte CONFIRMATION_SENT = 0x40;
	private final static byte CONFIRMATION_SW_ACKED = (byte) 0x80;
	
	private final static int DEFAULT_RETRY_SECONDS = 5;
	private final static int DEFAULT_ACK_RETRIES = 5;
	private final static int DEFAULT_CONFIRMATION_RETRIES = 5;
	private static int ackRetries = DEFAULT_ACK_RETRIES;
	private static int confirmationRetries = DEFAULT_CONFIRMATION_RETRIES;
	public static boolean recievedConfirmationAck = false;

	public static Map<Integer, AlarmMsg> temperatureAlarmsFlowState = new HashMap<Integer, AlarmMsg>();

	public static void handleAlarm(AlarmMsg alarm) {
		
		byte type = (byte) alarm.get_type();
		
		// Apliquem aquesta màscara perque de moment només ens interessen els
		// 4 bits de menor pes.
		switch(type & 0x0F) {
		case BATTERY:
			handleBatteryAlarm(alarm, type);
			break;
		case TEMPERATURE_AUTOMATIC:
			handleTemperatureAlarm(alarm, type);
			break;
		case TEMPERATURE_MANUAL:
			MessageLogger.showTemperatureAlarm(alarm, "manual");
			break;
		default:
			System.out.println("Alarma desconeguda:" + alarm.toString());
		}
		
	}
	
	private static void handleBatteryAlarm(AlarmMsg alarm, byte type) {
		
		// Apliquem aquesta màscara perque ara ens interessen els
		// 4 bits de major pes. Aquests bits ens indiquen quin es
		// l'estat de confirmació dels diferents punts de l'alarma.
		byte state = (byte) (type & 0xF0);
		
		switch(state) {
		// Estat SENT: Primer estat de l'alarma. La mota notifica a l'estació
		// base que ha detectat una batería baixa. 
		case ALARM_SENT:
			// En aquest cas només es requereix confirmació per software, 
			// per tant formatejem el missatge com a SW_ACK i l'enviem a la
			// mota que el va originar.
			alarm.set_type((short) (alarm.get_type() | ALARM_SW_ACKED));
			sendAlarmAck(alarm.get_moteId(), alarm);
			break;
		default:
			System.out.println("Alarma desconeguda:" + alarm.toString());
		}
		
	}
	
	private static void handleTemperatureAlarm(AlarmMsg alarm, byte type) {
		
		// Apliquem aquesta màscara perque ara ens interessen els
		// 4 bits de major pes. Aquests bits ens indiquen quin es
		// l'estat de confirmació dels diferents punts de l'alarma.
		byte state = (byte) (type & 0xF0);

		switch(state) {
		
		// Estat SENT: Primer estat de l'alarma. La mota notifica a l'estació
		// base que ha detectat una anomalía en la temperatura.
		case ALARM_SENT:
			// Es tracta d'una alarma nova. Comprovem si existeix en el sistema.
			if(temperatureAlarmsFlowState.containsKey(alarm.get_moteId())) {
				System.out.println("Ha entrat una alarma duplicada!");
			} 
			// Si no existeix l'afegim al sistema i n'enviem l'ACK a la mota.
			else {
				MessageLogger.showTemperatureAlarm(alarm, "automàtica");
				temperatureAlarmsFlowState.put(alarm.get_moteId(), alarm);
				alarm.set_type((short) (alarm.get_type() | ALARM_SW_ACKED));
				sendAlarmAck(alarm.get_moteId(), alarm);
			}
			break;
		// Estat USR_ACK: Últim estat de l'alarma. La mota notifica a l'estació
		// base que ha rebut correctament la confirmació manual de l'usuari.
		case (ALARM_SENT | ALARM_SW_ACKED | CONFIRMATION_SENT | CONFIRMATION_SW_ACKED):
			// Eliminem l'alarma del sistema ja que la tenim completament confirmada.
			recievedConfirmationAck = true;
			if(temperatureAlarmsFlowState.containsKey(alarm.get_moteId())) {
				temperatureAlarmsFlowState.remove(alarm.get_moteId());
			}
			System.out.println("La mota " + alarm.get_moteId() + " ha rebut correctament la confirmació de l\'usuari.");
			break;
		default:
			System.out.println("Alarma desconeguda:" + alarm.toString());
		}
		
	}
	
	private static void sendAlarmAck(int moteId, AlarmMsg msg) {
		try {
			for(int i=0; i<DEFAULT_ACK_RETRIES; i++) {
				Thread.sleep(1000);
				FDMonitor.instance.amSender.send(moteId, msg);
			}
			// Si es una alarma de temperatura, la guardem
			if((msg.get_type() | TEMPERATURE) == TEMPERATURE) {
				temperatureAlarmsFlowState.put(moteId, msg);
			}
			
		} catch (IOException e) {
			if(ackRetries > 0) {
				ackRetries--;
				sendAlarmAck(moteId, msg);
			} else {
				temperatureAlarmsFlowState.remove(moteId);
			}
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
	}
	
	public static void sendConfirmation(int moteId) {
		try {
			AlarmMsg msg = new AlarmMsg();
			msg.set_type((short) (ALARM_SENT | ALARM_SW_ACKED | CONFIRMATION_SENT));
			FDConfig.instance.amSender.send(moteId, msg);
			waitConfirmationAck(moteId, msg);
		} catch (IOException e) {
			if(confirmationRetries > 0) {
				System.out.println("No s'ha pogut enviar la configuració. " +
						"Esperant el reenviament.");
				confirmationRetries--;
				sendConfirmation(moteId);
			} else {
				recievedConfirmationAck = false;
				confirmationRetries = DEFAULT_CONFIRMATION_RETRIES;
			}
		} catch (InterruptedException e) {
			System.out.println("Error en temps d\'espera: " + e.getMessage());
		}
	}
	
	public static void waitConfirmationAck(int moteId, AlarmMsg msg) throws InterruptedException {
		for(int i=0; i<DEFAULT_RETRY_SECONDS; i++) {
			Thread.sleep(1000);
			System.out.print(".");
			if(recievedConfirmationAck) {
				break;
			}
		}
		System.out.println("");
		if(!recievedConfirmationAck) {
			confirmationRetries--;
			if(confirmationRetries>=0)
				sendConfirmation(moteId);
		} else {
			confirmationRetries = DEFAULT_CONFIRMATION_RETRIES;
			recievedConfirmationAck = false;
		}
	}
	
}
